package sample;

import java.awt.Container;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.UnknownHostException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.List;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.Preferences;
import java.lang.StringBuilder;

import javax.security.auth.login.LoginException;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.SwingUtilities;
import javax.swing.WindowConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.JFileChooser;
import javax.swing.SwingConstants;

import org.xml.sax.SAXException;

import sample.SwingBulkTable.SymbolCell;

import com.factset.rt.ControlHandler;
import com.factset.rt.MessageHandler;
import com.factset.rt.NotifyHandler;
import com.factset.rt.RT_Consumer;
import com.factset.rt.RT_Consumer.Subscription;
import com.factset.rt.RT_Consumer.BulkSubscription;
import com.factset.rt.RT_FidField;
import com.factset.rt.RT_FieldMap;
import com.factset.rt.RT_Message;
import com.factset.rt.RT_Request;

public class SwingBulk implements MessageHandler, ControlHandler, NotifyHandler {
	
	private JFrame frame;
	private JTabbedPane uriTabPane;
	private JComboBox enterpriseUriCombo;
	private JComboBox workstationUriCombo;
	private JComboBox serviceCombo;
	private JComboBox keyCombo;
	private JButton connectButton;
	private JButton disconnectButton;
	private JButton loadFileButton;
	private JButton requestButton;
	private JButton cancelButton;
	private JButton closeButton;
	private JCheckBox otpCheck;
	private JFileChooser fileChooser = new JFileChooser();
	private JLabel statusLabel;
	private JLabel updatesLabel;
	private ImageIcon blankIcon;
	private ImageIcon greenIcon;
	private ImageIcon greyIcon;
	private SwingBulkTable msgTable;
	
	private RT_Consumer consumer;
	private RT_FieldMap map;
	private Subscription requestedSub = null;
	private HashMap<Subscription, Integer> subRows = null;
	private final Logger logger = Logger.getLogger("SwingBulk");
	private final Preferences prefs = Preferences.userNodeForPackage(getClass());
	private int numUpdates = 0;
	
	private final Runnable dispatchRunnable = new Runnable() {
		@Override
		public void run() {
			consumer.dispatch();
		}
	};
	
	@Override
	public void onNotify() {
		SwingUtilities.invokeLater(dispatchRunnable);
	}
	
	@Override
	public void onMessage(Subscription sub, RT_Message msg) {
		int row = subRows.get(sub);
		if (msg.isError()) {
			sub.cancel();
			msgTable.setRow(row, msg.getKey(), blankIcon, "Invalid");
			updatesLabel.setText("Stream Closed: " + msg.getErrorDescription());
		}
		else if (msg.isStale()) {
			updatesLabel.setText("Stream Stale: " + msg.getStaleReason());
		}
		else {
			updatesLabel.setText("Stream OK, NumUpdates: " + numUpdates);
			
			String msgType = "", ask = "", bid = "";

			if (msg.isResponse()) {
				msgTable.setRow(row, msg.getKey(), blankIcon, "Pending");
			}
				
			statusLabel.setText("Key: " + msg.getKey() + "  NumFids: " +
					msg.count() + "  PermID: " + msg.getPermissions());

			for (RT_FidField field : msg) {
				String fieldName;
				if (map != null)
					fieldName = map.getFieldInfo(field.getFid()).getName();
				else
					fieldName = "{NOMAP}";
				if (fieldName.equals("MSG_TYPE")) {
					msgType = field.getValue();
				} else if (fieldName.equals("ASK_1")) {
					ask = field.getValue();
				} else if (fieldName.equals("BID_1")) {
					bid = field.getValue();
				}
			}
			ImageIcon icon = blankIcon;
			SymbolCell symCell = msgTable.getSymbolCellAt(row, 0);
			if (symCell.icon == greenIcon) {
				icon = greyIcon;
			} else {
				icon = greenIcon;
			}
			msgTable.setRow(row, symCell.symbol, icon, msgType);
			if (!ask.isEmpty()) {
				msgTable.setCell(row, 2, ask);
			}
			if (!bid.isEmpty()) {
				msgTable.setCell(row, 3, bid);
			}
			++numUpdates;
		}
	}
	
	@Override
	public void onControl(boolean isConnected, RT_Message controlMsg) {
		if (controlMsg.getKey().equals("CONNECTED")) {
			String[] services = consumer.getServices();
			for (int i=0; i<services.length; ++i)
				serviceCombo.addItem(services[i]);
			
			loadFileButton.setEnabled(true);
			requestButton.setEnabled(true);
			statusLabel.setText("Connected to: " +
					(String)((JComboBox)uriTabPane.getSelectedComponent()).getSelectedItem());
		}
		else if (controlMsg.getKey().equals("DISCONNECTED")) {
			serviceCombo.removeAllItems();
			loadFileButton.setEnabled(false);
			requestButton.setEnabled(false);
			statusLabel.setText("Lost connection... Retrying");
		}
		else if (controlMsg.getKey().equals("SERVICE_ENABLE")) {
			for (RT_FidField field : controlMsg) {
				serviceCombo.addItem(field.getValue());
			}
		}
		else if (controlMsg.getKey().equals("SERVICE_DISCONNECT")) {
			for (RT_FidField field : controlMsg) {
				serviceCombo.removeItem(field.getValue());
			}
		}
	}
	
	private void connect() {
		String uri =
				(String)((JComboBox)uriTabPane.getSelectedComponent()).getSelectedItem();
		if (uri == null)
			uri = "";
		
		try {
			if (uriTabPane.getSelectedComponent() == enterpriseUriCombo) {
				if(otpCheck.isSelected()) {
					JOptionPane.showMessageDialog(frame, "Please select a file named <keyId>.data and containing <your_key> on the first line and <your_counter> in decimal format on the second.", "Select Key/Counter File", JOptionPane.INFORMATION_MESSAGE);
					JFileChooser chooser = new JFileChooser();
					FileNameExtensionFilter filter = new FileNameExtensionFilter("Data Files (*.data)", "data");
					chooser.setFileFilter(filter);
					chooser.setCurrentDirectory(new File(System.getProperty("user.dir")));
					if(chooser.showOpenDialog(frame)!=JFileChooser.APPROVE_OPTION) {
						return;
					}
					String fullPath = chooser.getSelectedFile().getAbsolutePath();
					String path = "";
					String deviceId = "";
					int sepLoc = fullPath.lastIndexOf("\\");
					if(sepLoc!=-1) {
						path = fullPath.substring(0, sepLoc);
						deviceId = fullPath.substring(sepLoc+1);
					} else {
						deviceId = fullPath;
					}
					int dotLoc = deviceId.lastIndexOf(".");
					if(dotLoc!=-1) {
						deviceId = deviceId.substring(0, dotLoc);
					}
					String user = "";
					String host = "";
					int atLoc = uri.indexOf("@");
					if(atLoc!=-1) {
						user = uri.substring(0, atLoc);
						host = uri.substring(atLoc+1);
					}
					int colLoc = user.indexOf(":");
					if(colLoc!=-1) {
						user = user.substring(0, colLoc);
					}
					if(host.isEmpty() || user.isEmpty()) {
						JOptionPane.showMessageDialog(frame, "Host or user cannot be discerned. Ensure connection uri is in the form <user>-<serial>@<host>", "Error", JOptionPane.ERROR_MESSAGE);
						return;
					}
					if(deviceId.isEmpty()) {
						JOptionPane.showMessageDialog(frame, "Error determining keyId from file. Ensure key/counter file is named <keyId>.data", "Error", JOptionPane.ERROR_MESSAGE);
						return;
					}
					consumer = new RT_Consumer(host, user, deviceId, null, null, path, false);
				} else {
					consumer = new RT_Consumer(uri);
				}
			}
			else {
				String user;
				String pass = "";
				int delimIdx = uri.indexOf(':');
				if (delimIdx == -1)
					user = uri;
				else {
					user = uri.substring(0, delimIdx);
					pass = uri.substring(delimIdx+1);
				}
				consumer = RT_Consumer.newWorkstationInstance(user, pass);
			}
			consumer.setNotifyHandler(this);
			consumer.setControlHandler(this);
			logger.info("Connecting to " + uri);
			consumer.connect();
		} catch (UnknownHostException e) {
			String errMsg;
			if (uriTabPane.getSelectedComponent() == enterpriseUriCombo)
				errMsg = "Error resolving host from the connection string.";
			else
				errMsg = "No running workstation detected.";
				
			JOptionPane.showMessageDialog(frame, errMsg, "Error",
					JOptionPane.ERROR_MESSAGE);
			return;
		} catch (IOException e) {
			logger.severe("Error connecting to DataFeed server: " +
					e.getMessage());
			JOptionPane.showMessageDialog(frame, "Error connecting:\n" +
					e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
			return;
		} catch (LoginException e) {
			JOptionPane.showMessageDialog(frame,
					"Incorrect username or password.", "Error",
					JOptionPane.ERROR_MESSAGE);
			return;
		}
		
		addItemToCombo((JComboBox)uriTabPane.getSelectedComponent(), uri);
		uriTabPane.getSelectedComponent().setEnabled(false);
		uriTabPane.setEnabledAt((uriTabPane.getSelectedIndex() == 0 ? 1 : 0), false);
		
		for (String service : consumer.getServices())
			serviceCombo.addItem(service);
		serviceCombo.setSelectedItem(prefs.get("SERVICE", ""));
		
		connectButton.setEnabled(false);
		disconnectButton.setEnabled(true);
		loadFileButton.setEnabled(true);
		requestButton.setEnabled(true);
		frame.getRootPane().setDefaultButton(requestButton);
		
		statusLabel.setText("Connected to: " + uri);
	}
	
	private void disconnect() {
		consumer.disconnect();
		
		resetGUIState();
		requestedSub = null;
		numUpdates = 0;
	}
	
	private void loadFile() {
		int returnVal = fileChooser.showOpenDialog(frame);
		
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			File file = fileChooser.getSelectedFile();
			Path path = file.toPath();
			try {
				List<String> lines = Files.readAllLines(path, StandardCharsets.US_ASCII);
				StringBuilder sbContent = new StringBuilder();
				for (String s : lines) {
					sbContent.append(s);
					sbContent.append(',');
				}
				String contents = sbContent.toString();
				request(contents);
			} catch (IOException e) {
				JOptionPane.showMessageDialog(frame, "Error loading file " +
						file.toString() + ": " + e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
			}
		}
	}
	
	private void request(String key) {
		cancel();
		String service = (String)serviceCombo.getSelectedItem();
		prefs.put("SERVICE", service);
		
		RT_Request req = new RT_Request(service, key);		
		requestedSub = consumer.request(req, this);
		
		subRows = new HashMap<Subscription, Integer>();
		if (req.isExtendedRequest()) {
			BulkSubscription bulkSub = (BulkSubscription) requestedSub;
			List<Subscription> memberSubs = bulkSub.getSubscriptions();
			msgTable.setRowCount(memberSubs.size());
			int row = 0;
			for (Subscription sub : memberSubs) {
				if (subRows.get(sub) == null) {
					subRows.put(sub, row);
				}
				++row;
			}
		} else {
			msgTable.setRowCount(1);
			subRows.put(requestedSub, 0);
		}
		
		requestButton.setEnabled(false);
		loadFileButton.setEnabled(false);
		cancelButton.setEnabled(true);
		statusLabel.setText("Request sent for " + service + "::" + key);
	}
	
	private void cancel() {
		if (requestedSub != null) {
			requestedSub.cancel();
			requestedSub = null;
			
			requestButton.setEnabled(true);
			loadFileButton.setEnabled(true);
			cancelButton.setEnabled(false);
			numUpdates = 0;
			
			statusLabel.setText("Cancelled subscription");
			updatesLabel.setText("NumUpdates: 0");
		}
	}
	
	private void resetGUIState() {
		serviceCombo.removeAllItems();
		uriTabPane.getSelectedComponent().setEnabled(true);
		uriTabPane.setEnabled(true);
		uriTabPane.setEnabledAt((uriTabPane.getSelectedIndex() == 0 ? 1 : 0), true);
		
		connectButton.setEnabled(true);
		disconnectButton.setEnabled(false);
		loadFileButton.setEnabled(false);
		requestButton.setEnabled(false);
		cancelButton.setEnabled(false);
		
		statusLabel.setText("Disconnected");
		updatesLabel.setText("NumUpdates: 0");
	}
	
	private void addComponentsToPane() {
		Container pane = frame.getContentPane();
		((JComponent)pane).setBorder(BorderFactory.createEmptyBorder(0, 10,
				                                                     10, 10));
		pane.setLayout(new GridBagLayout());
		
        enterpriseUriCombo = new JComboBox();
        enterpriseUriCombo.setPreferredSize(new Dimension(225, 25));
        enterpriseUriCombo.setMinimumSize(new Dimension(225, 25));
        enterpriseUriCombo.setEditable(true);
        getItemsFromPrefs(enterpriseUriCombo, "ENT_URI", 3);
        
        workstationUriCombo = new JComboBox();
        workstationUriCombo.setEditable(true);
        getItemsFromPrefs(workstationUriCombo, "WKS_URI", 3);
        
        GridBagConstraints uriCon = new GridBagConstraints();
		uriCon.fill = GridBagConstraints.HORIZONTAL;
        uriCon.anchor = GridBagConstraints.LINE_START;
		uriCon.gridx = 0;
		uriCon.gridy = 0;
		uriCon.gridwidth = 3;
		uriCon.weightx = 0.08;
		uriCon.insets = new Insets(0,0,5,0);
		uriTabPane = new JTabbedPane();
		uriTabPane.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
		uriTabPane.addTab("Enterprise", enterpriseUriCombo);
		uriTabPane.addTab("Workstation", workstationUriCombo);
		uriTabPane.setSelectedIndex(prefs.getInt("TAB", 0));
		uriTabPane.addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent e) {
				if(uriTabPane.getSelectedComponent() == enterpriseUriCombo) {
					otpCheck.setEnabled(true);
				} else {
					otpCheck.setEnabled(false);
					otpCheck.setSelected(false);
				}
			}
		});
        pane.add(uriTabPane, uriCon);
        
        GridBagConstraints serviceCon = new GridBagConstraints();
        serviceCon.fill = GridBagConstraints.HORIZONTAL;
        serviceCon.gridx = 0;
        serviceCon.gridy = 1;
        serviceCombo = new JComboBox();
        pane.add(serviceCombo, serviceCon);
        
        GridBagConstraints keyCon = new GridBagConstraints();
        keyCon.fill = GridBagConstraints.HORIZONTAL;
        keyCon.anchor = GridBagConstraints.LINE_START;
        keyCon.gridx = 1;
        keyCon.gridy = 1;
        keyCon.gridwidth = 2;
        keyCon.insets = new Insets(0,5,0,0);
        keyCombo = new JComboBox();
        keyCombo.setEditable(true);
        getItemsFromPrefs(keyCombo, "KEY", 3);
        pane.add(keyCombo, keyCon);
        
        GridBagConstraints connectCon = new GridBagConstraints();
        connectCon.anchor = GridBagConstraints.LINE_END;
        connectCon.gridx = 3;
        connectCon.gridy = 0;
        connectCon.weightx = 0.1;
        connectCon.insets = new Insets(15,20,0,0);
        connectButton = new JButton();
        connectButton.setText("Connect");
        connectButton.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent e) {
        		connect();
        	}
        });
        frame.getRootPane().setDefaultButton(connectButton);
        pane.add(connectButton, connectCon);
        
        GridBagConstraints disconnectCon = new GridBagConstraints();
        disconnectCon.anchor = GridBagConstraints.LINE_START;
        disconnectCon.gridx = 4;
        disconnectCon.gridy = 0;
        disconnectCon.weightx = 0.5;
        disconnectCon.insets = new Insets(15,15,0,0);
        disconnectButton = new JButton();
        disconnectButton.setText("Disconnect");
        disconnectButton.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent e) {
        		disconnect();
        	}
        });
        pane.add(disconnectButton, disconnectCon);
        
        GridBagConstraints loadFileCon = new GridBagConstraints();
        loadFileCon.anchor = GridBagConstraints.LINE_END;
        loadFileCon.gridx = 3;
        loadFileCon.gridy = 2;
        loadFileCon.insets = new Insets(7,20,0,0);
        loadFileButton = new JButton();
        loadFileButton.setText("Load File");
        loadFileButton.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent e) {
        		loadFile();
        	}
        });
        pane.add(loadFileButton, loadFileCon);
        
        GridBagConstraints requestCon = new GridBagConstraints();
        requestCon.anchor = GridBagConstraints.LINE_END;
        requestCon.gridx = 3;
        requestCon.gridy = 1;
        requestCon.insets = new Insets(0,20,0,0);
        requestButton = new JButton();
        requestButton.setText("Request");
        requestButton.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent e) {
        		String key = (String)keyCombo.getSelectedItem();
        		if (key == null)
        			key = "";
        		
        		addItemToCombo(keyCombo, key);
        		request(key);
        	}
        });
        pane.add(requestButton, requestCon);
        
        GridBagConstraints cancelCon = new GridBagConstraints();
        cancelCon.anchor = GridBagConstraints.LINE_START;
        cancelCon.gridx = 4;
        cancelCon.gridy = 1;
        cancelCon.ipadx = 25;
        cancelCon.insets = new Insets(0,15,0,0);
        cancelButton = new JButton();
        cancelButton.setText("Cancel");
        cancelButton.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent e) {
        		cancel();
        	}
        });
        pane.add(cancelButton, cancelCon);

        GridBagConstraints otpCon = new GridBagConstraints();
        otpCon.insets = new Insets(10, 10, 0, 0);
        otpCon.anchor = GridBagConstraints.FIRST_LINE_START;
        otpCon.fill = GridBagConstraints.HORIZONTAL;
        otpCon.gridx = 4;
        otpCon.gridy = 2;
        otpCheck = new JCheckBox("Authenticate with OTP", false);
        otpCheck.setToolTipText("check for OTP authentication");
        pane.add(otpCheck, otpCon);

        msgTable = new SwingBulkTable(); 
        JScrollPane scrollPane = new JScrollPane(msgTable);

        GridBagConstraints tableCon = new GridBagConstraints();
        tableCon.fill = GridBagConstraints.BOTH;
        tableCon.gridx = 0;
        tableCon.gridy = 3;
        tableCon.ipady = 180;
        tableCon.gridwidth = 5;
        tableCon.weightx = 0.1;
        tableCon.weighty = 0.1;
        tableCon.insets = new Insets(10,0,0,0);
        pane.add(scrollPane, tableCon);
        
        GridBagConstraints statusCon = new GridBagConstraints();
        statusCon.fill = GridBagConstraints.HORIZONTAL;
        statusCon.gridx = 0;
        statusCon.gridy = 3;
        statusCon.gridwidth = 5;
        statusLabel = new JLabel();
        pane.add(statusLabel, statusCon);
        
        GridBagConstraints updatesCon = new GridBagConstraints();
        updatesCon.fill = GridBagConstraints.HORIZONTAL;
        updatesCon.gridx = 0;
        updatesCon.gridy = 4;
        updatesCon.gridwidth = 3;
        updatesLabel = new JLabel();
        pane.add(updatesLabel, updatesCon);
        
        GridBagConstraints closeCon = new GridBagConstraints();
        closeCon.anchor = GridBagConstraints.LINE_END;
        closeCon.insets = new Insets(0,0,10,0);
        closeCon.gridx = 4;
        closeCon.gridy = 4;
        closeButton = new JButton();
        closeButton.setText("Close");
        closeButton.addActionListener(new ActionListener() {
        	public void actionPerformed(ActionEvent e) {
        		frame.dispose();
        	}
        });
        
        resetGUIState();
        pane.add(closeButton, closeCon);
	}
	
	private static void addItemToCombo(JComboBox combo, String item) {
		for (int i = 0; i < combo.getItemCount(); ++i) {
			if (combo.getItemAt(i).equals(item)) {
				combo.removeItemAt(i);
				break;
			}
		}
		combo.insertItemAt(item, 0);
		combo.setSelectedIndex(0);
	}
	
	private void getItemsFromPrefs(JComboBox combo, String key, int numItems) {
		for (int i = 0; i < numItems; ++i) {
			String item = prefs.get(key + "_" + i, "");
			if (!item.isEmpty())
				combo.addItem(prefs.get(key + "_" + i, ""));
		}
	}
	
	private void putItemsInPrefs(JComboBox combo, String key, int numItems) {
		for (int i = 0; (i < numItems && i < combo.getItemCount()); ++i) {
			prefs.put(key + "_" + i, (String) combo.getItemAt(i));
		}
	}
	
	private ImageIcon loadIcon(String name) {
		URL iconURL = getClass().getResource(name);
		if (iconURL == null) {
			logger.warning("Error loading icon: " + name);
			return new ImageIcon();
		}
		else
			return new ImageIcon(iconURL);
	}
	
	private RT_FieldMap loadFieldMap(String uri) {
		try {
			return new RT_FieldMap(uri);
		} catch (SAXException e) {
			logger.log(Level.WARNING, "Error parsing field map", e);
			JOptionPane.showMessageDialog(frame,
					"Error parsing field map file: " + uri);
			return null;
		} catch (IOException e) {
			logger.log(Level.WARNING, "Error opening field map", e);
			JOptionPane.showMessageDialog(frame,
					"Error opening field map file: " + uri);
			return null;
		}
	}

	private void createAndShowGUI(String[] args) {	
		blankIcon = loadIcon("/resource/blank.gif");
		greenIcon = loadIcon("/resource/green.gif");
		greyIcon  = loadIcon("/resource/grey.gif");
		
        // Create and set up the window.
        frame = new JFrame("SwingBulk");
        frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
        frame.addWindowListener(new WindowAdapter() {
        	@Override
			public void windowClosed(WindowEvent we) {
        		if (consumer != null)
        			consumer.disconnect();
        		putItemsInPrefs(enterpriseUriCombo, "ENT_URI", 3);
        		putItemsInPrefs(workstationUriCombo, "WKS_URI", 3);
        		putItemsInPrefs(keyCombo, "KEY", 3);
        		prefs.putInt("TAB", uriTabPane.getSelectedIndex());
        	}
        });
        addComponentsToPane();
        
        // Display the window.
        frame.pack();
        frame.setSize(510, 800);
        frame.setVisible(true);
        
        try {
			FileHandler fh = new FileHandler("SwingBulk.log");
			logger.addHandler(fh);
			Logger.getLogger("com.factset.rt").addHandler(fh);
		} catch (IOException e) {
			logger.log(Level.WARNING, "Error opening log file", e);
		}
		
		if (args.length > 0)
			map = loadFieldMap(args[0]);
		else
			map = loadFieldMap("etc/rt_fields.xml");
    }

	public static void main(final String[] args) {
		// Schedule a job for the event-dispatching thread:
        // creating and showing this application's GUI.
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
            	SwingBulk main = new SwingBulk();
                main.createAndShowGUI(args);
            }
        });
	}
}
