package sample;

import javax.swing.Icon;
import javax.swing.JTable;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;

public class SwingQuoteTable extends JTable {

	private static final long serialVersionUID = 1L;
	
	private final DefaultTableModel tableModel;
	
	public static class FidCell implements Comparable<FidCell> {
		public Icon icon;
		public int fid;
		
		public FidCell(Icon icon, int fid) {
			this.icon = icon;
			this.fid  = fid;
		}

		@Override
		public int compareTo(FidCell o) {
			return (this.fid < o.fid ? -1 : (this.fid == o.fid ? 0 : 1));
		}
	}
	
	private static class FidCellRenderer extends DefaultTableCellRenderer {
		private static final long serialVersionUID = 1L;

		@Override
		public void setValue(Object value) {
			FidCell fidCell = (FidCell)value;
			setIcon(fidCell.icon);
			setText(Integer.toString(fidCell.fid));
		}
	}

	private static class SwingQuoteTableModel extends DefaultTableModel {
		private static final long serialVersionUID = 1L;

		@Override
		public Class<?> getColumnClass(int columnIndex) {
			if (columnIndex == 0)
				return FidCell.class;
			else
				return String.class;
		}
	}
	
	public SwingQuoteTable() {
		tableModel = new SwingQuoteTableModel();
		tableModel.setColumnIdentifiers(new String[] {"Fid", "Name", "Value"});
		
		super.setModel(tableModel);
		super.setDefaultRenderer(FidCell.class, new FidCellRenderer());
		super.setAutoCreateRowSorter(true);
		super.setEnabled(false);
		
		super.getColumnModel().getColumn(0).setPreferredWidth(70);
		super.getColumnModel().getColumn(1).setPreferredWidth(200);
		super.getColumnModel().getColumn(2).setPreferredWidth(280);
	}
	
	public FidCell getFidCellAt(int rowIndex, int columnIndex) {
		return (FidCell)tableModel.getValueAt(rowIndex, columnIndex);
	}
	
	@Override
	public void setValueAt(Object value, int rowIndex, int columnIndex) {
		tableModel.setValueAt(value, rowIndex, columnIndex);
	}
	
	public void setRow(int rowIndex, int fid, Icon icon, String fieldName,
			String fieldValue) {
		FidCell fidCell = new FidCell(icon, fid);
		
		tableModel.setValueAt(fidCell, rowIndex, 0);
		tableModel.setValueAt(fieldName, rowIndex, 1);
		tableModel.setValueAt(fieldValue, rowIndex, 2);
	}
	
	@Override
	public int getRowCount() {
		return tableModel.getRowCount();
	}
	
	public void setRowCount(int rowCount) {
		tableModel.setRowCount(rowCount);
	}
}
